use anyhow::Result;
use galactica_content::Content;
use glyphon::TextArea;
use log::{debug, error, trace};
use rhai::{Array, Dynamic, Engine, Scope, AST};
use std::{collections::HashSet, rc::Rc};

use super::{
	api::{self, SceneAction, SpriteElement, TextBoxBuilder},
	mouseevent::MouseEvent,
	util::{FpsIndicator, RadialBar, TextBox},
	UiElement, UiScene,
};
use crate::{
	ui::{
		api::{RadialBuilder, SpriteBuilder, State},
		util::Sprite,
	},
	RenderInput, RenderState,
};

pub(crate) struct UiManager {
	current_scene: UiScene,
	engine: Engine,
	scope: Scope<'static>,
	elements: Vec<UiElement>,
	ct: Rc<Content>,

	show_timings: bool,
	fps_indicator: FpsIndicator,
}

impl UiManager {
	pub fn new(ct: Rc<Content>, state: &mut RenderState) -> Self {
		let scope = Scope::new();

		let mut engine = Engine::new();
		api::register_into_engine(&mut engine);

		Self {
			ct,
			current_scene: UiScene::Flying,
			engine,
			scope,
			elements: Vec::new(),
			show_timings: true,
			fps_indicator: FpsIndicator::new(state),
		}
	}

	pub fn get_scene_ast(ct: &Content, scene: UiScene) -> &AST {
		match scene {
			UiScene::Landed => &ct.get_config().ui_landed_scene,
			UiScene::Flying => &ct.get_config().ui_flying_scene,
			UiScene::Outfitter => &ct.get_config().ui_outfitter_scene,
		}
	}

	/// Change the current scene
	pub fn set_scene(&mut self, state: &mut RenderState, scene: UiScene) -> Result<()> {
		debug!("switching to {:?}", scene);
		self.current_scene = scene;

		self.scope.clear();
		self.elements.clear();
		let mut used_names = HashSet::new();

		trace!("running init for `{}`", self.current_scene.to_string());

		let builders: Array = self.engine.call_fn(
			&mut self.scope,
			Self::get_scene_ast(&self.ct, self.current_scene),
			"init",
			(State {
				planet_landscape: "ui::landscape::test".to_string(),
				planet_name: "Earth".to_string(),
			},),
		)?;

		trace!("found {:?} builders", builders.len());

		for v in builders {
			if v.is::<SpriteBuilder>() {
				let s = v.cast::<SpriteBuilder>();
				if used_names.contains(s.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.to_string(),
						s.name
					);
				} else {
					used_names.insert(s.name.to_string());
				}
				self.elements.push(UiElement::new_sprite(Sprite::new(
					&self.ct,
					s.name.to_string(),
					s.sprite.to_string(),
					s.mask.map(|x| x.to_string()),
					s.rect,
				)));
			} else if v.is::<RadialBuilder>() {
				let r = v.cast::<RadialBuilder>();
				if used_names.contains(r.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.to_string(),
						r.name
					);
				} else {
					used_names.insert(r.name.to_string());
				}
				self.elements.push(UiElement::new_radialbar(RadialBar::new(
					&self.ct,
					r.name.to_string(),
					r.stroke,
					r.color,
					r.rect,
					r.progress,
				)));
			} else if v.is::<TextBoxBuilder>() {
				let t = v.cast::<TextBoxBuilder>();
				if used_names.contains(t.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.to_string(),
						t.name
					);
				} else {
					used_names.insert(t.name.to_string());
				}
				let mut b = TextBox::new(
					state,
					t.name.to_string(),
					t.font_size,
					t.line_height,
					t.font,
					t.justify,
					t.rect,
				);
				b.set_text(state, &t.text);
				self.elements.push(UiElement::new_text(b));
			} else {
				// TODO: better message
				error!("bad type in builder array")
			}
		}
		return Ok(());
	}

	/// Draw all ui elements
	pub fn draw(&mut self, input: &RenderInput, state: &mut RenderState) -> Result<()> {
		let mut iter = self.elements.iter();

		if self.show_timings {
			self.fps_indicator.step(input, state);
		}

		let action: SceneAction = loop {
			let e = match iter.next() {
				Some(e) => e,
				None => break SceneAction::None,
			};

			let action: Dynamic = {
				if let Some(e) = e.sprite() {
					let mut x = (*e).borrow_mut();
					let m = x.check_mouse(input, state);
					x.step(input, state);
					x.push_to_buffer(input, state);
					drop(x);
					// we MUST drop here, since script calls mutate the sprite RefCell

					match m {
						MouseEvent::None => Dynamic::from(SceneAction::None),

						MouseEvent::Release | MouseEvent::Click => self.engine.call_fn(
							&mut self.scope,
							Self::get_scene_ast(&self.ct, self.current_scene),
							"click",
							(SpriteElement::new(self.ct.clone(), e.clone()), m.is_click()),
						)?,

						MouseEvent::Leave | MouseEvent::Enter => self.engine.call_fn(
							&mut self.scope,
							Self::get_scene_ast(&self.ct, self.current_scene),
							"hover",
							(SpriteElement::new(self.ct.clone(), e.clone()), m.is_enter()),
						)?,
					}
				} else if let Some(e) = e.radialbar() {
					let mut x = (*e).borrow_mut();
					x.step(input, state);
					x.push_to_buffer(input, state);
					Dynamic::from(SceneAction::None)
				} else {
					Dynamic::from(SceneAction::None)
				}
			};

			if let Some(action) = action.try_cast::<SceneAction>() {
				match action {
					SceneAction::None => {}
					_ => {
						break action;
					}
				}
			}
		};

		drop(iter);

		match action {
			SceneAction::None => {}
			SceneAction::SceneOutfitter => self.set_scene(state, UiScene::Outfitter)?,
			SceneAction::SceneLanded => self.set_scene(state, UiScene::Landed)?,
		}

		return Ok(());
	}
}

impl<'a> UiManager {
	/// Get textareas
	pub fn get_textareas(
		&'a mut self,
		input: &RenderInput,
		state: &RenderState,
	) -> Vec<TextArea<'a>> {
		let mut v = Vec::with_capacity(32);
		if self.show_timings {
			v.push(self.fps_indicator.get_textarea(state, input))
		}

		for t in self
			.elements
			.iter()
			.map(|x| x.text())
			.filter(|x| x.is_some())
			.map(|x| x.unwrap())
			.map(|x| x.get_textarea(state, input))
		{
			v.push(t)
		}

		v
	}
}
